@@ -7,7 +7,7 @@ use DatumType::F32;
 fn mat_vec_mul(c: &mut Criterion) {
     let mut group = c.benchmark_group("mat_vec_mul");
     unsafe {
-        for (m, k) in [(64usize, 64usize)].iter() {
+        for (m, k) in [(768usize, 256usize)].iter() {
             group.throughput(Throughput::Elements((m * k) as u64));
             group.bench_with_input(
                 BenchmarkId::from_parameter(format!("{}x{}", m, k)),
@@ -16,8 +16,8 @@ fn mat_vec_mul(c: &mut Criterion) {
                     let mm =
                         tract_linalg::ops().mmm(F32, F32, F32, Some(m), Some(k), Some(1)).unwrap();
                     let pa = Tensor::uninitialized_aligned::<f32>(
-                        &[mm.a_pack(k).len(m)],
-                        mm.a_pack(k).alignment(),
+                        &[mm.a_pack().len(k, m)],
+                        mm.a_pack().alignment(),
                     )
                     .unwrap();
                     let b = tensor1(&vec![0.0; k]);
@@ -29,10 +29,10 @@ fn mat_vec_mul(c: &mut Criterion) {
                             &[
                                 FusedSpec::AddMatMul {
                                     a: mm.a_packed(F32.size_of(), k).wrap(&pa.view()),
-                                    b: mm.b_packed(b.datum_type().size_of(), k).wrap(&b.view()),
+                                    b: mm.b_packed(b.datum_type().size_of(), k).wrap(&b.view()).unwrap(),
                                     k,
                                 },
-                                FusedSpec::Store(mm.c_view().wrap(&c.view_mut())),
+                                FusedSpec::Store(mm.c_view(0, 0).wrap(&c.view_mut())),
                             ],
                         )
                     });
