@@ -1,8 +1,9 @@
 use std::{env, fs};
-mod armv7neon;
+pub mod armv7neon;
 mod armvfpv2;
 use crate::frame::ElementWiseImpl;
-use crate::frame::MatMatMulImpl;
+
+use crate::frame::mmm::kernel::MatMatMulKer;
 
 use crate::Ops;
 
@@ -37,57 +38,47 @@ pub fn plug(ops: &mut Ops) {
         log::info!("armv7neon activated (smmm, ssigmoid), stanh)");
         let cpu = cpu_part().unwrap_or(0);
 
-        fn prefer_8x4(_m: Option<usize>, _k: Option<usize>, n:Option<usize>) -> bool {
+        fn prefer_8x4(_m: Option<usize>, _k: Option<usize>, n: Option<usize>) -> bool {
             n.map(|n| n % 4 == 0 && n % 6 != 0 && n <= 12).unwrap_or(false)
         }
 
         ops.mmv_f32 = match cpu {
-            0xc07 => Box::new(|_, _| {
-                Box::new(MatMatMulImpl::<armv7neon::MatMatMulF32x32x1CortexA7, f32>::new())
-            }),
-            0xc09 => Box::new(|_, _| {
-                Box::new(MatMatMulImpl::<armv7neon::MatMatMulF32x32x1CortexA9, f32>::new())
-            }),
-            _ => Box::new(|_, _| {
-                Box::new(MatMatMulImpl::<armv7neon::MatMatMulF32x32x1Generic, f32>::new())
-            }),
+            0xc07 => Box::new(|_, _| armv7neon::armv7neon_mmm_f32_32x1_cortexa7::mmm()),
+            0xc09 => Box::new(|_, _| armv7neon::armv7neon_mmm_f32_32x1_cortexa9::mmm()),
+            _ => Box::new(|_, _| armv7neon::armv7neon_mmm_f32_32x1_generic::mmm()),
         };
 
         ops.mmm_f32 = match cpu {
-            0xc07 => Box::new(|m, k, n| {
+            0xc07 => Some(Box::new(|m, k, n| {
                 if prefer_8x4(m, k, n) {
-                    Box::new(MatMatMulImpl::<armv7neon::MatMatMulF32x8x4CortexA7, f32>::new())
+                    armv7neon::armv7neon_mmm_f32_8x4_cortexa7::mmm()
                 } else {
-                    Box::new(MatMatMulImpl::<armv7neon::MatMatMulF32x8x6CortexA7, f32>::new())
+                    armv7neon::armv7neon_mmm_f32_8x6_cortexa7::mmm()
                 }
-            }),
-            0xc09 => Box::new(|m, k, n| {
+            })),
+            0xc09 => Some(Box::new(|m, k, n| {
                 if prefer_8x4(m, k, n) {
-                    Box::new(MatMatMulImpl::<armv7neon::MatMatMulF32x8x4CortexA9, f32>::new())
+                    armv7neon::armv7neon_mmm_f32_8x4_cortexa9::mmm()
                 } else {
-                    Box::new(MatMatMulImpl::<armv7neon::MatMatMulF32x8x6CortexA9, f32>::new())
+                    armv7neon::armv7neon_mmm_f32_8x6_cortexa9::mmm()
                 }
-            }),
-            _ => Box::new(|m, k, n| {
+            })),
+            _ => Some(Box::new(|m, k, n| {
                 if prefer_8x4(m, k, n) {
-                    Box::new(MatMatMulImpl::<armv7neon::MatMatMulF32x8x4Generic, f32>::new())
+                    armv7neon::armv7neon_mmm_f32_8x4_generic::mmm()
                 } else {
-                    Box::new(MatMatMulImpl::<armv7neon::MatMatMulF32x8x6Generic, f32>::new())
+                    armv7neon::armv7neon_mmm_f32_8x6_generic::mmm()
                 }
-            }),
+            })),
         };
-        ops.qmmm_i32 =
-            Box::new(|_, _, _| Box::new(MatMatMulImpl::<armv7neon::MatMatMulI8x8x4, i32>::new()));
-        ops.qmmv_i32 =
-            Box::new(|_, _| Box::new(MatMatMulImpl::<armv7neon::MatMatMulI8x32x1, i32>::new()));
+        ops.qmmm_i32 = Box::new(|_, _, _| armv7neon::armv7neon_mmm_i32_8x4::mmm());
+        ops.qmmv_i32 = Box::new(|_, _| armv7neon::armv7neon_mmm_i32_32x1::mmm());
         ops.sigmoid_f32 =
             Box::new(|| Box::new(ElementWiseImpl::<armv7neon::SigmoidF32x4n, f32>::new()));
         ops.tanh_f32 = Box::new(|| Box::new(ElementWiseImpl::<armv7neon::TanhF32x4n, f32>::new()));
-        ops.prefetch = Some(&armv7neon::prefetch);
     } else {
         log::info!("armvfpv2 activated for smmm");
-        ops.mmm_f32 =
-            Box::new(|_, _, _| Box::new(MatMatMulImpl::<armvfpv2::MatMatMulF32x4x4, f32>::new()));
+        ops.mmm_f32 = Some(Box::new(|_, _, _| armvfpv2::armvfpv2_mmm_f32_4x4::mmm()));
     }
 }
 
