@@ -1,47 +1,25 @@
-use num_traits::{AsPrimitive, Zero};
+use num_traits::AsPrimitive;
 use std::marker::PhantomData;
 use std::{fmt, ops};
 
 use tract_data::prelude::*;
 
 use super::*;
-use crate::frame::mmm::InputStoreKer::*;
 use crate::frame::mmm::*;
-
-use num_traits::sign::Signed;
+use crate::LADatum;
 
 #[derive(Copy, Clone, Debug)]
 pub struct GenericMmm4x4<TA, TB, TI>(PhantomData<(TA, TB, TI)>)
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + ScaleShiftAndRound
-        + PartialOrd
-        + Zero
-        + fmt::Debug
-        + fmt::Display
-        + 'static;
+    TI: LADatum + ScaleShiftAndRound;
 
 unsafe impl<TA, TB, TI> Send for GenericMmm4x4<TA, TB, TI>
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + ScaleShiftAndRound
-        + PartialOrd
-        + Zero
-        + fmt::Debug
-        + fmt::Display
-        + 'static,
+    TI: LADatum + ScaleShiftAndRound,
 {
 }
 
@@ -49,17 +27,7 @@ unsafe impl<TA, TB, TI> Sync for GenericMmm4x4<TA, TB, TI>
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + ScaleShiftAndRound
-        + PartialOrd
-        + Zero
-        + fmt::Debug
-        + fmt::Display
-        + 'static,
+    TI: LADatum + ScaleShiftAndRound,
 {
 }
 
@@ -97,23 +65,16 @@ impl<TA, TB, TI> MatMatMulKer<TI> for GenericMmm4x4<TA, TB, TI>
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + ScaleShiftAndRound
-        + PartialOrd
-        + Zero
-        + Signed
-        + fmt::Debug
-        + fmt::Display
-        + 'static,
+    TI: LADatum + ScaleShiftAndRound,
     usize: AsPrimitive<TI>,
 {
     #[inline(always)]
     fn name() -> &'static str {
-        "generic"
+        match TI::datum_type() {
+            DatumType::F32 => "generic_f32_4x4",
+            DatumType::I32 => "generic_i32_4x4",
+            _ => panic!(),
+        }
     }
     #[inline(always)]
     fn mr() -> usize {
@@ -184,62 +145,26 @@ where
                     }
                     FusedKerSpec::AddMatMul { k, pa, pb, .. } => {
                         let a = pa as *const TA;
-                        match *pb {
-                            Packed(PackedStoreKer { ptr: b }) => {
-                                let b = b as *const TB;
-                                for i in 0..k {
-                                    let a = std::slice::from_raw_parts(a.offset(4 * i as isize), 4);
-                                    let b = std::slice::from_raw_parts(b.offset(4 * i as isize), 4);
-                                    ab[0][0] += a[0].as_() * b[0].as_();
-                                    ab[0][1] += a[0].as_() * b[1].as_();
-                                    ab[0][2] += a[0].as_() * b[2].as_();
-                                    ab[0][3] += a[0].as_() * b[3].as_();
-                                    ab[1][0] += a[1].as_() * b[0].as_();
-                                    ab[1][1] += a[1].as_() * b[1].as_();
-                                    ab[1][2] += a[1].as_() * b[2].as_();
-                                    ab[1][3] += a[1].as_() * b[3].as_();
-                                    ab[2][0] += a[2].as_() * b[0].as_();
-                                    ab[2][1] += a[2].as_() * b[1].as_();
-                                    ab[2][2] += a[2].as_() * b[2].as_();
-                                    ab[2][3] += a[2].as_() * b[3].as_();
-                                    ab[3][0] += a[3].as_() * b[0].as_();
-                                    ab[3][1] += a[3].as_() * b[1].as_();
-                                    ab[3][2] += a[3].as_() * b[2].as_();
-                                    ab[3][3] += a[3].as_() * b[3].as_();
-                                }
-                            }
-                            OffsetsAndPtrs { row_byte_offsets, col_ptrs } => {
-                                let col_ptrs = col_ptrs as *const *const TB;
-                                let pb0 = *(col_ptrs.offset(0));
-                                let pb1 = *(col_ptrs.offset(1));
-                                let pb2 = *(col_ptrs.offset(2));
-                                let pb3 = *(col_ptrs.offset(3));
-                                for i in 0..k {
-                                    let a = std::slice::from_raw_parts(a.offset(4 * i as isize), 4);
-                                    let offset = *row_byte_offsets.offset(i as isize)
-                                        / std::mem::size_of::<TB>() as isize;
-                                    let b0 = *(pb0.offset(offset));
-                                    let b1 = *(pb1.offset(offset));
-                                    let b2 = *(pb2.offset(offset));
-                                    let b3 = *(pb3.offset(offset));
-                                    ab[0][0] += a[0].as_() * b0.as_();
-                                    ab[0][1] += a[0].as_() * b1.as_();
-                                    ab[0][2] += a[0].as_() * b2.as_();
-                                    ab[0][3] += a[0].as_() * b3.as_();
-                                    ab[1][0] += a[1].as_() * b0.as_();
-                                    ab[1][1] += a[1].as_() * b1.as_();
-                                    ab[1][2] += a[1].as_() * b2.as_();
-                                    ab[1][3] += a[1].as_() * b3.as_();
-                                    ab[2][0] += a[2].as_() * b0.as_();
-                                    ab[2][1] += a[2].as_() * b1.as_();
-                                    ab[2][2] += a[2].as_() * b2.as_();
-                                    ab[2][3] += a[2].as_() * b3.as_();
-                                    ab[3][0] += a[3].as_() * b0.as_();
-                                    ab[3][1] += a[3].as_() * b1.as_();
-                                    ab[3][2] += a[3].as_() * b2.as_();
-                                    ab[3][3] += a[3].as_() * b3.as_();
-                                }
-                            }
+                        let b = pb as *const TB;
+                        for i in 0..k {
+                            let a = std::slice::from_raw_parts(a.offset(4 * i as isize), 4);
+                            let b = std::slice::from_raw_parts(b.offset(4 * i as isize), 4);
+                            ab[0][0] += a[0].as_() * b[0].as_();
+                            ab[0][1] += a[0].as_() * b[1].as_();
+                            ab[0][2] += a[0].as_() * b[2].as_();
+                            ab[0][3] += a[0].as_() * b[3].as_();
+                            ab[1][0] += a[1].as_() * b[0].as_();
+                            ab[1][1] += a[1].as_() * b[1].as_();
+                            ab[1][2] += a[1].as_() * b[2].as_();
+                            ab[1][3] += a[1].as_() * b[3].as_();
+                            ab[2][0] += a[2].as_() * b[0].as_();
+                            ab[2][1] += a[2].as_() * b[1].as_();
+                            ab[2][2] += a[2].as_() * b[2].as_();
+                            ab[2][3] += a[2].as_() * b[3].as_();
+                            ab[3][0] += a[3].as_() * b[0].as_();
+                            ab[3][1] += a[3].as_() * b[1].as_();
+                            ab[3][2] += a[3].as_() * b[2].as_();
+                            ab[3][3] += a[3].as_() * b[3].as_();
                         }
                     }
                     FusedKerSpec::Store(tile) => store(&tile, &ab),
@@ -256,33 +181,13 @@ pub struct GenericMmm4x1<TA, TB, TI>(PhantomData<(TA, TB, TI)>)
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + ScaleShiftAndRound
-        + PartialOrd
-        + Zero
-        + fmt::Debug
-        + fmt::Display
-        + 'static;
+    TI: LADatum + ScaleShiftAndRound;
 
 unsafe impl<TA, TB, TI> Send for GenericMmm4x1<TA, TB, TI>
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + ScaleShiftAndRound
-        + PartialOrd
-        + Zero
-        + fmt::Debug
-        + fmt::Display
-        + 'static,
+    TI: LADatum + ScaleShiftAndRound,
 {
 }
 
@@ -290,17 +195,7 @@ unsafe impl<TA, TB, TI> Sync for GenericMmm4x1<TA, TB, TI>
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + ScaleShiftAndRound
-        + PartialOrd
-        + Zero
-        + fmt::Debug
-        + fmt::Display
-        + 'static,
+    TI: LADatum + ScaleShiftAndRound,
 {
 }
 
@@ -308,23 +203,16 @@ impl<TA, TB, TI> MatMatMulKer<TI> for GenericMmm4x1<TA, TB, TI>
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + ScaleShiftAndRound
-        + PartialOrd
-        + Zero
-        + Signed
-        + fmt::Debug
-        + fmt::Display
-        + 'static,
+    TI: LADatum + ScaleShiftAndRound,
     usize: AsPrimitive<TI>,
 {
     #[inline(always)]
     fn name() -> &'static str {
-        "generic"
+        match TI::datum_type() {
+            DatumType::F32 => "generic_f32_4x1",
+            DatumType::I32 => "generic_i32_4x1",
+            _ => panic!(),
+        }
     }
     #[inline(always)]
     fn mr() -> usize {
@@ -400,32 +288,14 @@ where
                     }
                     FusedKerSpec::AddMatMul { k, pa, pb, .. } => {
                         let a = pa as *const TA;
-                        match *pb {
-                            Packed(PackedStoreKer { ptr: b }) => {
-                                let b = b as *const TB;
-                                for i in 0..k {
-                                    let a = std::slice::from_raw_parts(a.offset(4 * i as isize), 4);
-                                    let b = *b.offset(i as isize);
-                                    ab[0][0] += a[0].as_() * b.as_();
-                                    ab[1][0] += a[1].as_() * b.as_();
-                                    ab[2][0] += a[2].as_() * b.as_();
-                                    ab[3][0] += a[3].as_() * b.as_();
-                                }
-                            }
-                            OffsetsAndPtrs { row_byte_offsets, col_ptrs } => {
-                                let col_ptrs = col_ptrs as *const *const TB;
-                                let pb0 = *(col_ptrs.offset(0));
-                                for i in 0..k {
-                                    let a = std::slice::from_raw_parts(a.offset(4 * i as isize), 4);
-                                    let offset = *row_byte_offsets.offset(i as isize)
-                                        / std::mem::size_of::<TB>() as isize;
-                                    let b0 = *(pb0.offset(offset));
-                                    ab[0][0] += a[0].as_() * b0.as_();
-                                    ab[1][0] += a[1].as_() * b0.as_();
-                                    ab[2][0] += a[2].as_() * b0.as_();
-                                    ab[3][0] += a[3].as_() * b0.as_();
-                                }
-                            }
+                        let b = pb as *const TB;
+                        for i in 0..k {
+                            let a = std::slice::from_raw_parts(a.offset(4 * i as isize), 4);
+                            let b = *b.offset(i as isize);
+                            ab[0][0] += a[0].as_() * b.as_();
+                            ab[1][0] += a[1].as_() * b.as_();
+                            ab[2][0] += a[2].as_() * b.as_();
+                            ab[3][0] += a[3].as_() * b.as_();
                         }
                     }
                     FusedKerSpec::Store(tile) => store(
@@ -451,32 +321,14 @@ pub struct GenericMmmTest3x2<TA, TB, TI>(PhantomData<(TA, TB, TI)>)
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + PartialOrd
-        + Zero
-        + fmt::Debug
-        + fmt::Display
-        + 'static;
+    TI: LADatum + ScaleShiftAndRound;
 
 #[cfg(test)]
 unsafe impl<TA, TB, TI> Send for GenericMmmTest3x2<TA, TB, TI>
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + PartialOrd
-        + Zero
-        + fmt::Debug
-        + fmt::Display
-        + 'static,
+    TI: LADatum + ScaleShiftAndRound,
 {
 }
 
@@ -485,16 +337,7 @@ unsafe impl<TA, TB, TI> Sync for GenericMmmTest3x2<TA, TB, TI>
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + PartialOrd
-        + Zero
-        + fmt::Debug
-        + fmt::Display
-        + 'static,
+    TI: LADatum + ScaleShiftAndRound,
 {
 }
 
@@ -503,23 +346,16 @@ impl<TA, TB, TI> MatMatMulKer<TI> for GenericMmmTest3x2<TA, TB, TI>
 where
     TA: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
     TB: Datum + Copy + fmt::Debug + AsPrimitive<TI>,
-    TI: Datum
-        + Copy
-        + ops::AddAssign
-        + ops::Mul<Output = TI>
-        + ops::MulAssign
-        + PartialOrd
-        + ScaleShiftAndRound
-        + Zero
-        + Signed
-        + fmt::Debug
-        + fmt::Display
-        + 'static,
+    TI: LADatum + ScaleShiftAndRound,
     usize: AsPrimitive<TI>,
 {
     #[inline(always)]
     fn name() -> &'static str {
-        "generic-test-3x2"
+        match TI::datum_type() {
+            DatumType::F32 => "generic_f32_3x2",
+            DatumType::I32 => "generic_i32_3x2",
+            _ => panic!(),
+        }
     }
     #[inline(always)]
     fn mr() -> usize {
@@ -590,38 +426,16 @@ where
                     }
                     FusedKerSpec::AddMatMul { k, pa, pb, .. } => {
                         let a = pa as *const TA;
-                        match *pb {
-                            Packed(PackedStoreKer { ptr: b }) => {
-                                let b = b as *const TB;
-                                for i in 0..k {
-                                    let a = std::slice::from_raw_parts(a.offset(3 * i as isize), 3);
-                                    let b = std::slice::from_raw_parts(b.offset(2 * i as isize), 2);
-                                    ab[0][0] += a[0].as_() * b[0].as_();
-                                    ab[0][1] += a[0].as_() * b[1].as_();
-                                    ab[1][0] += a[1].as_() * b[0].as_();
-                                    ab[1][1] += a[1].as_() * b[1].as_();
-                                    ab[2][0] += a[2].as_() * b[0].as_();
-                                    ab[2][1] += a[2].as_() * b[1].as_();
-                                }
-                            }
-                            OffsetsAndPtrs { row_byte_offsets, col_ptrs } => {
-                                let col_ptrs = col_ptrs as *const *const TB;
-                                let pb0 = *(col_ptrs.offset(0));
-                                let pb1 = *(col_ptrs.offset(1));
-                                for i in 0..k {
-                                    let a = std::slice::from_raw_parts(a.offset(3 * i as isize), 3);
-                                    let offset = *row_byte_offsets.offset(i as isize)
-                                        / std::mem::size_of::<TB>() as isize;
-                                    let b0 = *(pb0.offset(offset));
-                                    let b1 = *(pb1.offset(offset));
-                                    ab[0][0] += a[0].as_() * b0.as_();
-                                    ab[0][1] += a[0].as_() * b1.as_();
-                                    ab[1][0] += a[1].as_() * b0.as_();
-                                    ab[1][1] += a[1].as_() * b1.as_();
-                                    ab[2][0] += a[2].as_() * b0.as_();
-                                    ab[2][1] += a[2].as_() * b1.as_();
-                                }
-                            }
+                        let b = pb as *const TB;
+                        for i in 0..k {
+                            let a = std::slice::from_raw_parts(a.offset(3 * i as isize), 3);
+                            let b = std::slice::from_raw_parts(b.offset(2 * i as isize), 2);
+                            ab[0][0] += a[0].as_() * b[0].as_();
+                            ab[0][1] += a[0].as_() * b[1].as_();
+                            ab[1][0] += a[1].as_() * b[0].as_();
+                            ab[1][1] += a[1].as_() * b[1].as_();
+                            ab[2][0] += a[2].as_() * b[0].as_();
+                            ab[2][1] += a[2].as_() * b[1].as_();
                         }
                     }
                     FusedKerSpec::Store(tile) => store(&tile, &ab),
@@ -663,7 +477,7 @@ where
 
 unsafe fn add_unicast<TI, AB>(tile: &OutputStoreKer, ab: &mut [AB])
 where
-    TI: Datum + ops::AddAssign<TI> + Copy,
+    TI: LADatum + ops::AddAssign<TI>,
     AB: AsMut<[TI]> + fmt::Debug,
 {
     if tile.item_size == TI::datum_type().size_of() {
@@ -692,11 +506,32 @@ where
     }
 }
 
-test_mmm_kernel_f32!(crate::generic::mmm::GenericMmm4x4<f32, f32, f32>, test_GenericMmm4x4_f32, true);
-test_mmm_kernel_i8!(crate::generic::mmm::GenericMmm4x4<i8, i8, i32>, test_GenericMmm4x4_i8, true);
+#[cfg(test)]
+#[allow(non_camel_case_types)]
+type generic_f32_4x4 = GenericMmm4x4<f32, f32, f32>;
+test_mmm_kernel_f32!(generic_f32_4x4, true);
 
-test_mmm_kernel_f32!(crate::generic::mmm::GenericMmm4x1<f32, f32, f32>, test_GenericMmm4x1_f32, true);
-test_mmm_kernel_i8!(crate::generic::mmm::GenericMmm4x1<i8, i8, i32>, test_GenericMmm4x1_i8, true);
+#[cfg(test)]
+#[allow(non_camel_case_types)]
+type generic_i32_4x4 = GenericMmm4x4<i8, i8, i32>;
+test_mmm_kernel_i32!(generic_i32_4x4, true);
 
-test_mmm_kernel_f32!(crate::generic::mmm::GenericMmmTest3x2<f32, f32, f32>, test_GenericMmmTest3x2_f32, true);
-test_mmm_kernel_i8!(crate::generic::mmm::GenericMmmTest3x2<i8, i8, i32>, test_GenericMmmTest3x2_i8, true);
+#[cfg(test)]
+#[allow(non_camel_case_types)]
+type generic_f32_4x1 = GenericMmm4x1<f32, f32, f32>;
+test_mmm_kernel_f32!(generic_f32_4x1, true);
+
+#[cfg(test)]
+#[allow(non_camel_case_types)]
+type generic_i32_4x1 = GenericMmm4x1<i8, i8, i32>;
+test_mmm_kernel_i32!(generic_i32_4x1, true);
+
+#[cfg(test)]
+#[allow(non_camel_case_types)]
+type generic_f32_3x2 = GenericMmmTest3x2<f32, f32, f32>;
+test_mmm_kernel_f32!(generic_f32_3x2, true);
+
+#[cfg(test)]
+#[allow(non_camel_case_types)]
+type generic_i32_3x2 = GenericMmmTest3x2<i8, i8, i32>;
+test_mmm_kernel_i32!(generic_i32_3x2, true);
